package uk.co.vault101.ritual;

import static com.badlogic.gdx.scenes.scene2d.actions.Actions.fadeOut;
import static com.badlogic.gdx.scenes.scene2d.actions.Actions.run;

import com.badlogic.gdx.Game;
import com.badlogic.gdx.Screen;
import com.badlogic.gdx.assets.AssetManager;
import com.badlogic.gdx.graphics.g2d.BitmapFont;
import com.badlogic.gdx.graphics.g2d.GlyphLayout;
import com.badlogic.gdx.graphics.g3d.Material;
import com.badlogic.gdx.scenes.scene2d.Stage;
import com.badlogic.gdx.scenes.scene2d.actions.SequenceAction;
import com.badlogic.gdx.utils.Array;

/**
 * With a tip from "Game Programming Patterns" (check if title correct)
 * this helps with cohesion, to disassociate graphics/sounds/ai/physics
 * @author Pict
 *
 */
public abstract class GraphicalComponent {

	protected Ritual display;
	protected GlyphLayout glyphLayout = new GlyphLayout();
	protected Stage stage;
	protected BitmapFont font;

	public AssetManager assets;
	public WorldMap map;
	public Material[] previewMaterial;
	public Tile[] podiums;

	/** Permanent expected instances */
	public final Array<GameObject> instances = new Array<>();
	/** Dynamic, temporary unplanned instances */
	public final Array<GameObject> tempInstances = new Array<>();
	
	
	public GraphicalComponent(final Ritual display) {
		this.display = display;
	}
	
	/**
	 * Hold a reference to the Screen (or state)
	 * that uses this Graphical component
	 * 
	 * Note: A generic 'Screen' is not held in this class as it's expected 
	 * that implementing classes hold their own more specific (narrower)
	 * class reference. e.g. hold 'GameState' instead of Screen.
	 * 
	 * @param screen
	 */
	public abstract void setScreenRef(final Screen screen);
		
	public void render (final float delta) {
		update3D (delta);
		render3D (delta);
		update2D (delta);
		render2D (delta);
	}

	protected abstract void update3D(float delta);
	protected abstract void render3D(float delta);
	
	protected abstract void update2D(float delta);
	protected abstract void render2D(float delta);
	
	/** What to do with Graphics when a screen should be shown */
	public abstract void show();
	/** What to do with Graphics when a screen should be hidden */
	public abstract void hide();
	
	/** What to do with Graphics when a screen should be paused */
	public abstract void pause();
	/** What to do with Graphics when a screen should be resumed */
	public abstract void resume();
	
	/** What to do with Graphics when a screen should be resized */
	public abstract void resize(final int width, final int height);
	/** What to do with Graphics when a screen should be disposed */
	public abstract void dispose();
	
	public abstract void nextScreen();
	
	/**
	 * Define this once here instead of in EVERY state class
	 * @param game Existing game instance - needed for .setScreen(screen) call
	 * @param newScreen Screen instance to switch to
	 */
	public void switchScreen(final Game game, final Screen newScreen){
	    stage.getRoot().getColor().a = 1;
	    SequenceAction sequenceAction = new SequenceAction();
	    sequenceAction.addAction(fadeOut(0.5f));
	    sequenceAction.addAction(run(new Runnable() {
	        @Override
	        public void run() {
	            game.setScreen(newScreen);
	        }
	    }));
	    stage.getRoot().addAction(sequenceAction);
	}
}
