package uk.co.vault101.ritual;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import uk.co.vault101.ritual.model.GameItem;

import com.badlogic.gdx.assets.AssetManager;
import com.badlogic.gdx.graphics.g3d.Material;
import com.badlogic.gdx.graphics.g3d.Model;
import com.badlogic.gdx.math.Vector3;

public class Tile implements ToJSON {
	
	public final static Vector3 dimensions = new Vector3(1,0,1);

	private float x,y,z;
	/** Model instance for the architecture (tile/wall/etc) */
	private GameObject gameObject;
	/** Game item that can be picked up, or used for decoration */
	private GameItem gameItem;
	private boolean pickable;
	private int rotation;
	private int materialIndex = -1;
	private static final String FORMAT_1F = "%.1f";

	/** These must be unique and distinct. */
	public enum TileType {
		NORMAL, IMPASSABLE
	}
	
	private boolean impassable;
	
	// pathfinding stuff
	private boolean visited;
	private int cost;
    
	public Tile(final float x, final float y, final float z, final WorldMap mapRef) {
		this.x = x;
		this.y = y;
		this.z = z;
		this.cost = 1;
	}

	public GameObject getGameObject() {
		return gameObject;
	}
	
	public GameItem getGameItem() {
		return gameItem;
	}
	
	public int getMaterialIndex() {
		return materialIndex;
	}
	
    public int getRotation() {
    	return this.rotation;
    }
    
    public int getCost() {
    	return cost;
    }
    
    public float getX() {
		return x;
	}
    
    public float getY() {
		return y;
	}
    
    public float getZ() {
		return z;
	}
    
    public boolean isImpassable() {
    	return impassable;
    }

	public boolean isPickable() {
    	return pickable;
    }
	
	public boolean isVisited() {
		return visited;
	}

	public void rotateLeft() {
		rotation -= 90;
		if (rotation < 0) {
			rotation += 360;
		}
		rotation %= 360;
		getGameObject().transform.rotate(0, 1, 0, -90);
		getGameObject().recalculateBounds();
	}

	public void rotateRight() {
		rotation += 90;
		if (rotation < 0) {
			rotation += 360;
		}
		rotation %= 360;
		getGameObject().transform.rotate(0, 1, 0, 90);
		getGameObject().recalculateBounds();
	}

	public void setGameObject(final GameObject gameObject) {
		gameObject.tileRef = this;
		this.gameObject = gameObject;
	}
	

	public void setGameItem(final GameItem gameItem) {
		gameItem.getGameObject().tileRef = this;
		this.gameItem = gameItem;
		setTranslate(getX(), getY(), getZ());
	}
	
	public void setImpassable(final boolean impassable) {
		this.impassable = impassable;
	}
	
	public void setMaterialIndex(final int texturePointer) {
		this.materialIndex = texturePointer;
		cost = 1;
	}
	
	public void setPickable(final boolean pickable) {
		this.pickable = pickable;
	}
	
	public void setRotation(final int rotation) {
    	this.rotation = rotation;
    }
	
	public void setCost(final int cost) {
		this.cost = cost;
	}
	
	public void incrementCost() {
		this.cost++;
	}
	
	public void decrementCost() {
		if (this.cost >= 2) {
			this.cost--;
		}
	}

	public void setTranslate(final float x, final float y, final float z) {
    	this.x = x;
    	this.y = y;
    	this.z = z;
    	
    	gameObject.transform.setToTranslation(x,y,z);
    	if (gameItem != null) {
    		gameItem.getGameObject().transform.setToTranslation(x,y+(gameItem.isPodiumDisplayed() ? 0.5f : 0f), z);
    	}
    }
	
	public void setVisited(final boolean visited) {
		this.visited = visited;
	}
	
	public static void loadAll(final JSONObject jsonMap, final WorldMap mapRef, final AssetManager assets, final Material[] previewMaterial) throws JSONException {
		final JSONArray jsonTiles = jsonMap.getJSONArray("tiles");
		for (int i=0; i < jsonTiles.length(); i++) {
			final JSONObject jsonTile = jsonTiles.getJSONObject(i);
			final Tile tile = new Tile((float)jsonTile.getDouble("x"), (float)jsonTile.getDouble("y"), (float)jsonTile.getDouble("z"), mapRef);

			mapRef.setTile((int)tile.getX(), (int)tile.getZ(), tile);
			tile.setPickable(jsonTile.getBoolean("p"));
			tile.setImpassable(jsonTile.optBoolean("i", false));
			tile.setMaterialIndex(jsonTile.getInt("mat"));
			
			final JSONObject jsonGameObject = jsonTile.getJSONObject("g");
		
			final GameObject go = new GameObject(assets.get(jsonGameObject.getString("name"), Model.class), jsonGameObject.getString("name"));

			tile.setGameObject(go);
			if (tile.getMaterialIndex() > -1) {
				final Material mat = tile.getGameObject().materials.first();
				mat.clear();
				mat.set(previewMaterial[tile.getMaterialIndex()]);
			}
			
			tile.setTranslate(tile.getX(), tile.getY(), tile.getZ());
			// TODO: rather inefficient, possibly doing 3 rotations for one rotate right
			while (tile.getRotation() != jsonTile.getInt("rot")) {
				tile.rotateLeft();
			}
			mapRef.getInstances().add(go);
		} // for each Tile TODO move the load of a tile into the tile class
	}

	/** Used in saving the object, helps with debugging. Native/binary/compressed can wait */
	@Override
	public JSONObject toJSON() {
		final JSONObject jo = new JSONObject();
		try {
			jo.put("c", "Tile"); // class
			jo.put("x", String.format(FORMAT_1F, x)); // we only allow height diffs in 0.1 increments.
			jo.put("y", String.format(FORMAT_1F, y));
			jo.put("z", String.format(FORMAT_1F, z));
			jo.put("rot", rotation);
			jo.put("p", pickable);
			jo.put("i", impassable);
			jo.put("mat", materialIndex);
			jo.put("g", gameObject.toJSON());
		} catch (JSONException e) {};
		return jo;
	}
	
	/**
	 * debugging only, probably in the Console or logs
	 * Expect this won't contain as much information as toJSON();
	 */
	@Override
	public String toString() {
		final JSONObject jo = new JSONObject();
		try {
			jo.put("x", String.format(FORMAT_1F, x));
			jo.put("y", String.format(FORMAT_1F, y));
			jo.put("z", String.format(FORMAT_1F, z));
			jo.put("rot", rotation);
			jo.put("p", pickable);
			jo.put("i", impassable);
			jo.put("mat", materialIndex);
			jo.put("c", cost);
		} catch (JSONException e) {};
		return jo.toString();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + Float.floatToIntBits(x);
		result = prime * result + Float.floatToIntBits(y);
		result = prime * result + Float.floatToIntBits(z);
		return result;
	}

	@Override
	public boolean equals(final Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final Tile other = (Tile) obj;
		if (Float.floatToIntBits(x) != Float.floatToIntBits(other.x)) {
			return false;
		}
		if (Float.floatToIntBits(y) != Float.floatToIntBits(other.y)) {
			return false;
		}
		if (Float.floatToIntBits(z) != Float.floatToIntBits(other.z)) {
			return false;
		}
		return true;
	}

	public void clearGameItem() {
		gameItem = null;
	}
}
