package uk.co.vault101.ritual;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import uk.co.vault101.ritual.ai.NPC;
import uk.co.vault101.ritual.ai.pathfinding.AStarPathFinder;
import uk.co.vault101.ritual.ai.pathfinding.Mover;
import uk.co.vault101.ritual.ai.pathfinding.PathFinder;
import uk.co.vault101.ritual.ai.pathfinding.TileBasedMap;
import uk.co.vault101.ritual.model.GameItem;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.assets.AssetManager;
import com.badlogic.gdx.graphics.g3d.Material;
import com.badlogic.gdx.utils.Array;

public class WorldMap implements TileBasedMap, ToJSON {
	
	/** Maximum size of a map that is currently supported */
	public static final int MAX_WORLD_DIMENSION = 32;
	public static final int ROOM_SIZE = 9;
	/** 2D array of tiles */
	private final Tile[][] map;
	private final Array<GameObject> instances = new Array<>();
	/** becomes true if the map has had a change made to it, and a save could be required */
	private boolean dirtybit = false;
	private PathFinder pathfinder;
	private List<NPC> npcs = new ArrayList<>();
	
	public final String filename;
	public final int width, height;
	public final boolean generated;
	private final Random random = new Random();

	private List<GameItem> availableGameItems = new ArrayList<>();

	/** Create a new map of maximum dimensions */
	public WorldMap() {
		this.map = new Tile[MAX_WORLD_DIMENSION][MAX_WORLD_DIMENSION];
		this.width = MAX_WORLD_DIMENSION;
		this.height = MAX_WORLD_DIMENSION;
		this.filename = "maps/map_"+Long.toString(System.currentTimeMillis(), 36);
		this.generated = true;

	}
	
	/** Create a new map in the world of a particular width and height (TODO: probably unused) */
	public WorldMap(final int width, final int height) {
		this.map = new Tile[height][width];
		this.width = width;
		this.height = height;
		this.filename = "maps/map_"+Long.toString(System.currentTimeMillis(), 36);
		this.generated = true;
	}

	public Tile[][] getMap() {
		return map;
	}

	/** Loads a map from internal assets with the given name. "maps/" path prefix is expected and should be passed in
	 * @throws JSONException */
	public WorldMap(final String mapName, final AssetManager assets, final Material[] previewMaterial) throws JSONException {

		this.generated = false;

		if (!mapName.startsWith("maps/")) {
			// TODO: console?
			throw new IllegalArgumentException("Unexpected map name, must have maps/ prefix. Not writing to another area of files.internal");
		}
		this.filename = mapName;
		final JSONObject jsonMap = new JSONObject(Gdx.files.local(mapName).readString());

		// TODO: quick check we're processing the correct thing here. This could come out, to be honest
		if (!jsonMap.getString("c").equals(this.getClass().getSimpleName())) {
			throw new JSONException("Expected JSON Object to match class: "+this.getClass().getSimpleName()+" but is: "+jsonMap.getString("c"));
		}

		final int width = jsonMap.getInt("width");
		final int height = jsonMap.getInt("height");

		this.map = new Tile[width][height];
		this.width = width;
		this.height = height;

		Tile.loadAll(jsonMap, this, assets, previewMaterial);
	}

	public void setAvailableGameItems(List<GameItem> availableGameItems) {
		this.availableGameItems = availableGameItems;
	}

	public List<GameItem> getAvailableGameItems() {
		return availableGameItems;
	}

	/** Main update method for the entire map and everything in it
	 * 
	 * @param delta
	 */
	public void update (final float delta) {
		updateNPCs(delta);
	}
		
	@Override
	public boolean blocked(final Mover mover, final int x, final int y) {
		return getTile(x,y).isImpassable();
	}
	
	/**
	 * Clear which tiles have been visited by the path finder.
	 * Should only really be called when we create a new path
	 */
	public void clearVisited() {
		for (int x=0;x<getWidthInTiles();x++) {
			for (int y=0;y<getHeightInTiles();y++) {
				getTile(x,y).setVisited(false);
			}
		}
	}
	
	/**
	 * Return the A* pathfinder object for this map. If none exists, creates a new one
	 * @return PathFinder for A* path finding
	 */
	public PathFinder getAStarPathFinder(final boolean allowDiagonalMovement) {
		if (pathfinder == null) {
			pathfinder = new AStarPathFinder(this, 50, allowDiagonalMovement);
		}
		return pathfinder;
	}
	
	/**
	 * 
	 */
	@Override
	public float getCost(final Mover mover, final int sx, final int sy, final int tx, final int ty) {
		return getTile(sx, sy).getCost();
	}

	@Override
	public int getHeightInTiles() {
		return height;
	}
	
	public Array<GameObject> getInstances() {
		return instances;
	}
	
	public Tile getTile(final int x, final int y) {
		return map[y][x];
	}
	
	@Override
	public int getWidthInTiles() {
		return width;
	}
	
	public boolean isDirtyBit() {
		return this.dirtybit;
	}
	

	@Override
	public void pathFinderVisited(final int x, final int y) {
		getTile(x,y).setVisited(true);
	}

	public void setDirtybit() {
		this.dirtybit = true;
	}
	
	/**
	 * Save i.e. overwrite a map that was loaded
	 */
	public boolean save() {
		try {
			// Note: .local, not .internal as internal is read only!
			//MAR save to diff file each time
			//Gdx.files.local(filename).writeString(this.toJSON().toString(4), false);
			Gdx.files.local("maps/map_saved_" + System.currentTimeMillis()).writeString(this.toJSON().toString(4), false);

			dirtybit = false; // clear dirty bit
			return true;
		} catch (JSONException e) {
			System.err.println("Unable to save map to formatted json in "+filename+", dumping to output");
			System.err.println(e.getMessage());
			e.printStackTrace(System.err);
			try {
				System.out.println(this.toJSON().toString(4));
			} catch (JSONException e1) {
				System.err.println("Unable to save map to formatted json");
			}
			return false;
		} catch (Throwable t) {
			System.err.println(t.getMessage());
			t.printStackTrace(System.err);
			try {
				System.out.println(this.toJSON().toString(4));
			} catch (JSONException e1) {
				System.err.println("Unable to save map to formatted json");
			}
			return false;
		}
	}

	/** Return the tile passed in, to allow chaining */
	public Tile setTile(final int x, final int y, final Tile tile) {
		map[y][x] = tile;
		return tile;
	}

	/** Used in saving the object, helps with debugging. Native/binary/compressed can wait */
	@Override
	public JSONObject toJSON() {
		final JSONObject jo = new JSONObject();
		try {
			jo.put("c", "WorldMap") // class
			.put("width", width)
			.put("height", height);
			
			// Tiles
			final JSONArray tiles = new JSONArray();
			for (int y=0; y < height; y++) {
				for (int x=0; x < width; x++) {
					tiles.put(map[y][x].toJSON());
				}
			}
			jo.put("tiles", tiles);
			
		} catch (JSONException e) {};
		return jo;
	}
	
	/**
	 * Called before loading a new map to clear up some of the old junk.
	 */
	public void dispose() {
		if (map != null) {
			for (int y=0; y < map.length; y++) {
				for (int x=0; x < map[y].length; x++) {
					map[y][x] = null;
				}
			}
		}
	}

	public static WorldMap load(final String filename, final AssetManager assets, final Material[] previewMaterial) {
		final WorldMap map;
		try {
        	if (Gdx.files.local(filename).exists() && Gdx.files.local(filename).length() > 0) {
        		map = new WorldMap(filename, assets, previewMaterial);
        		return map;
        	} else {
            	map = new WorldMap();
            	return map;
        	}
        } catch (JSONException e) {
        	System.err.println("[WorldMap] Problem reading map: "+filename+" "+e.getMessage());
        }
		return null;
	}

	public Tile determineSpawnTile() {

		Tile theSpawnTile = null;
		Tile possibleSpawnTile;

		while (theSpawnTile == null) {
			final int x = random.nextInt(MAX_WORLD_DIMENSION);
			final int z = random.nextInt(MAX_WORLD_DIMENSION);
			possibleSpawnTile = map[z][x];

			if (possibleSpawnTile.isImpassable()) {
				// we don't want to know about it
			} else if (possibleSpawnTile.getGameItem() == null) { // not already a game item here
				theSpawnTile = possibleSpawnTile;
			}
		}

		return theSpawnTile;
	}

	public void initialiseGameItems(final GameItem... gameItems) {

		for (int i = 0; i < gameItems.length; i++) {

			// true for collectables, false for detail
			//final GameItem gameItem = new GameItem("BLAH!", new GameObject(gameItems[i], "Heid"), true);

			getAvailableGameItems().add(gameItems[i]);

			final Tile theTile = determineSpawnTile();

			theTile.setGameItem(gameItems[i]);
		}
	}
	
	/** Update all NPCs */
	private void updateNPCs(final float delta) {
		for (final NPC npc : npcs) {
			npc.update(delta, this);
		}
	}
	
	/** Add an NPC to the map */
	public NPC addNPC(final AssetManager assets, final String name, final float x, final float y, final float z, final GameState gs) {
		final NPC npc = new NPC(assets, x,y,z, name, this, gs);
		npcs.add(npc);
		
		instances.add(npc.getGameObject());
		return npc;
	}
	
	/** Get a list of NPCs */
	public List<NPC> getNPCs() {
		return npcs;
	}
}
